// Copyright (c) 2015, <your name>. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

/// The plot_window library.
///
/// This is an awesome library. More dartdocs go here.

library plot_window;

import "dart:html";
import "dart:math" as math;
import "dart:async";
import "package:intl/intl.dart" as intl;

part "line.dart";
part "axis.dart";
part "color.dart";
part "settings.dart";
part "point_list.dart";
part "point_smoother.dart";
part "plot_action.dart";
part "text_performable.dart";
part "legend_position.dart";
part "line_plot_window.dart";
part "heatmap_plot_window.dart";
part "heatmap.dart";

// TODO: Export any libraries intended for clients of this package.

void Init() {
  LegendPosition.Init();
  AverageMethod.Init();
  LineType.Init();
}


abstract class PlotWindow {

  CanvasElement canvas;
  CanvasRenderingContext2D context;
  bool multiPlot = false;

  LegendPosition legendPosition = LegendPosition.TOPRIGHT;
  bool gridOntop=false;
  bool showGrid=true;
  
  DataListSetting get legendPositionSetting=>new DataListSetting("Legend position", legendPosition, LegendPosition.LegendPositions, (String s) { legendPosition=LegendPosition.GetLegendPosition(s);});

  BoolSetting get gridOntopSetting=>new BoolSetting("Grid on top", gridOntop, (bool v) { gridOntop=v; });
  BoolSetting get showGridSetting=>new BoolSetting("Show grid", showGrid, (bool v) { showGrid=v; });
  SettingsGroup settings;

  UListElement settingsContainer = new UListElement();
  void _initSettingsContainer() {
    settingsContainer.style..position="fixed"
      ..display="none"
      ..width="60vw"
      ..height="40vh"
      ..marginTop="20vh"
      ..marginLeft="20vw"
      ..padding="5vw"
      ..top="0"
      ..left="0"
      ..backgroundColor="white"
      ..textAlign="middle"
      ..boxSizing="border-box"
      ..overflowY="auto";
  }
  
  void removeData();
  
  void showSettings() {
    settingsContainer.children..clear()
      ..add(settings.html);
    settingsContainer.style.display="block";
  }

  void hideSettings() {
    settingsContainer.style.display="none";
  }
  
  ColorIterator colorIterator = new ColorIterator();
  
  PlotActionList plotActionList = new PlotActionList();
  
  Map<AxisType, Axis> axes = {AxisType.Y:new Axis(), AxisType.X:new Axis(), AxisType.Y2:new Axis()};

  bool forceSquare = false;
  bool grid = false;
  bool legend = true;

  
  NumPerformable fontSize = new NumPerformable("enter font size", 10);
  String fontFace = "sans-serif";
  String getFontString()=>"${fontSize.value}px ${fontFace}";

  num longestYAxisText = 0;
  num get paddingLeft=>longestYAxisText + 2*fontSize.value + yAxisLabel.fontSize;
  num get paddingTop=>title.fontSize+2*fontSize.value;
  num get paddingBottom => xAxisLabel.fontSize + fontSize.value * 3;
  num get paddingRight => 0;
  num get graphWidth => canvas.width - paddingLeft -paddingRight;
  num get graphHeight => canvas.height - paddingTop - paddingBottom;

  TextPerformable xAxisLabel = new TextPerformable("x axis label");
  TextPerformable yAxisLabel = new TextPerformable("y axis label");

  Rectangle _rectangle;
  Rectangle get rectangle => _rectangle;

  void set rectangle(Rectangle r) {
    _rectangle=r;
    axes.values.forEach((Axis axis)=>axis.sizeFluid = false);
  }

  static Rectangle _shift(Rectangle r, Point offset) {
    return new Rectangle.fromPoints(r.topLeft-offset, r.bottomRight-offset);
  }

  static Rectangle _zoom(Rectangle r, num factor, [ List<AxisType> axisTypes, Point center]) {
    if (center==null)
      center = new Point(r.left + r.width/2, r.top + r.height/2);
    num left = r.left;
    num width = r.width;
    num top = r.top;
    num height = r.height;
    if (axisTypes==null || axisTypes.isEmpty || axisTypes.contains(AxisType.Y)) {
      top -= center.y;
      top *= factor;
      top += center.y;
      height *= factor;
    }
    if (axisTypes==null || axisTypes.isEmpty || axisTypes.contains(AxisType.X)) {
      left -= center.x;
      left *= factor;
      left += center.x;
      width *= factor;
    }
    return new Rectangle(left, top, width, height);
  }

  bool mouseDrag = false;
  Point oldOffset;
  bool _rendering = false;
  bool get rendering=>_rendering;
  void set rendering(bool r) {
    _rendering=r;
  }
  
  
  PlotWindow(this.canvas) {
    this.settings = new SettingsGroup("Plot Window Settings");
    _initSettingsContainer();
    this.canvas.parent.children.add(this.settingsContainer);
    this.context = canvas.context2D;
    this.canvas.onMouseWheel.listen((WheelEvent e) {
      if(this.rendering) return;
      this.rendering=true;
      axes.values.forEach((Axis axis)=>axis.sizeFluid = false);
      this.clear();
      List<AxisType> axisTypes = new List<AxisType>();
      num delta = e.deltaY;
      if (e.altKey)
        axisTypes.add(AxisType.Y);
      if (e.shiftKey) {
        axisTypes.add(AxisType.X);
        delta = e.deltaX;
      }
      var factor = 1 + (delta/1000)/math.sqrt(delta/1000 * delta/1000 + 1);
      this.rectangle = _zoom(this.rectangle, factor, axisTypes, toRectangle(e.offset));
      this.plot();
      this.rendering=false;
    });
    this.canvas.onMouseDown.listen((var e) {
      this.mouseDrag = true;
      this.oldOffset = e.offset;
    });
    this.canvas.onMouseUp.listen((_)=>this.mouseDrag = false);
    this.canvas.onMouseOut.listen((_)=>this.mouseDrag = false);
    this.canvas.onMouseMove.listen((var e) {
      if (this.mouseDrag) {
        if(this.rendering) return;
        this.rendering = true;
        var offset = toRectangle(e.offset) - toRectangle(this.oldOffset);
        this.rectangle = _shift(this.rectangle, offset);
        this.oldOffset = e.offset;
        plot();
        this.rendering=false;
      }
    });
    this.canvas.onClick.listen((MouseEvent e) {
      e.preventDefault();
      e.stopPropagation();
      plotActionList.getPlotAction(e.offset)?.action();
    });
  }

  /* accepts a point in x, y coordinates and returns the canvas coordinates */
  Point toCanvas(Point p) {
    num x = (p.x - rectangle.left)/rectangle.width * graphWidth + paddingLeft;
    num y = (1 - (p.y - rectangle.top)/rectangle.height) *graphHeight + paddingTop; 
    return new Point(x, y);
  }

  /* accepts a points in canvas coordinates and returns x,y values */
  Point toRectangle(Point p) {
    num x = (p.x - paddingLeft)/graphWidth * rectangle.width + rectangle.left;
    num y = (1 - (p.y - paddingTop)/graphHeight)*rectangle.height + rectangle.top; 
    return new Point(x, y);
  }
  
  void clear() {
    this.context.clearRect(0,0,canvas.width, canvas.height);
    this.plotActionList.clear();
  }

  List<Point> filterToRectangle(List<Point> points) {
    int firstIndex=0;
    int lastIndex=points.length-1;

    for (int i=0; i<points.length; i++) {
      if(rectangle.left<=points[i].x)
        break;
      firstIndex=i;
    }

    for (int i=points.length-1; i>0; i--) {
      if (rectangle.right>=points[i].x) //containsPoint(points[i]))
        break;
      lastIndex=i;
    }
    return points.getRange(firstIndex, lastIndex+1).toList();
  }

  Rectangle getInitialSize();
  
  void plot() {
    this.clear();
    Axis.TargetTicDistance = 10*fontSize.value;
    if(axes.values.map((Axis a)=>a.sizeFluid).contains(true) && !this.mouseDrag)
      rectangle = _zoom(getInitialSize(), 1.1);
    setYTics();
    setXTics();
    if(gridOntop) {
      drawContent();
      drawGrid(showGrid);
    }  else {
      drawGrid(showGrid);
      drawContent();
    }
    //    blankOutBorders();
    drawTics();
    drawAxisLabels();
    drawLegend();
    drawTitle();
  }

  void drawContent();
  
  TextPerformable title = new TextPerformable("enter title");
  
  void drawTitle() {
    context.save();
    context.beginPath();
    context.fillStyle = "${Color.WHITE}";
    context.strokeStyle = "${Color.BLACK}";
    num lineWidth = 1;
    context.lineWidth = lineWidth;
    context.font = title.fontString; 
    num length = context.measureText(title.value).width;
    num padding = fontSize.value/2;
    num height = fontSize.value + 2*padding;
    num left = canvas.width/2 - length/2 - padding;
    num width = length + 2 * padding;
    num top = fontSize.value;
    //    context.stroke();
    context.fillStyle = "${Color.BLACK}";
    context.textAlign = "center";
    context.textBaseline = "top";
    context.fillText(title.value, canvas.width/2, top);
    plotActionList.addPlotAction(new PlotAction(new Rectangle(left, top, width, height), () async {
      settings.populate(title, true);
      settings.populate(xAxisLabel, false);
      settings.populate(yAxisLabel, false);
      settings.populate(fontSize, false);
      settings.addPerformer(gridOntopSetting);
      settings.addPerformer(showGridSetting);
      showSettings();
      await settings.run();
      hideSettings();
      plot();
    }));
    context.restore();
  }

  List<LegendEntry> getLegendEntries();

  void drawLegend() {
    context.beginPath();
    context.fillStyle="${Color.WHITE}";
    context.strokeStyle = "${Color.BLACK}";
    var lineWidth=1;
    context.lineWidth=lineWidth;
    context.font = getFontString();
    List<LegendEntry> legendEntries = getLegendEntries();
    num maxLength = legendEntries.map((LegendEntry le)=>le.strs.map((String s)=>context.measureText(s).width).reduce((num val, num element)=>math.max(val, element))).reduce((num val, num element)=>math.max(val, element));
    num numEntries = legendEntries.map((LegendEntry le)=>le.strs.length).reduce((num current, num elem)=>current+elem);
    num padding = fontSize.value/2;
    num lineSpacing = fontSize.value+5;
    num left;
    num top;
    num width = maxLength+2*padding;
    num height = numEntries * lineSpacing + 2*padding;
    if(legendPosition.right)
      left=canvas.width-width-lineWidth-paddingRight;
    else
      left = paddingLeft;
    if(legendPosition.bottom)
      top=canvas.height - height - lineWidth - paddingBottom;
    else
      top = paddingTop;
    context.rect(left, top, width, height);
    
    context.fill();
    context.stroke();
    
    plotActionList.addPlotAction(new PlotAction(new Rectangle(left, top, width, height), () async {
      settings.clear();
      for (LegendEntry le in legendEntries) {
        settings.populate(le.performable);
      }
      settings.addPerformer(legendPositionSetting);
      showSettings();
      await settings.run();
      hideSettings();
      plot();
    }));
    
    context.textAlign = "start";
    left += padding;
    top += padding;
    for(LegendEntry le in legendEntries) { 
      var color = le.color;
      context.strokeStyle = "$color";
      context.fillStyle = "$color";
      for(String str in le.strs) {
        context.fillText(str, left, top + fontSize.value);
        top += lineSpacing;
      }
    }
  }

  
  void setXTics() { this.axes[AxisType.X].setTics(rectangle.left, rectangle.right, graphWidth); } //canvas.width); }

  void setYTics() {
    this.axes[AxisType.Y].setTics(rectangle.top, rectangle.bottom, graphHeight);
    context.font = getFontString();
    longestYAxisText=this.axes[AxisType.Y].tics.map((num yTic)=>context.measureText(yTic.toStringAsFixed(axes[AxisType.Y].fixedPrecision)).width).reduce(math.max);
  }

  void blankOutBorders() {
    context.save();
    context.fillStyle="${Color.WHITE}";
    context.fillRect(0,0, paddingLeft, canvas.height);
    context.fillRect(0,0, canvas.width, paddingTop);
    context.fillRect(0, canvas.height - paddingBottom, canvas.width, paddingBottom);
    context.fillRect(canvas.width-paddingRight, 0, paddingRight, canvas.height);
    context.restore();
  }
  
  void drawGrid(bool showLightLines) {
    context.save();
    context.strokeStyle = "${Color.BLACK}";
    var yTics = this.axes[AxisType.Y].tics;
    var xTics = this.axes[AxisType.X].tics;
    for (num yTic in yTics) {
      if (yTic!=0 && !showLightLines) continue;
      if (yTic==0) context.lineWidth = 2;
      else context.lineWidth = 0.5;
      context..beginPath()
        ..moveTo(paddingLeft,toCanvas(new Point(0,yTic)).y)
        ..lineTo(canvas.width-paddingRight, toCanvas(new Point(0,yTic)).y)
        ..stroke();
    }
    
    for(num xTic in xTics) {
      if (xTic!=0 &&!showLightLines) continue;
      if (xTic==0) context.lineWidth = 2;
      else context.lineWidth = 0.5;
      context..beginPath()
        ..moveTo(toCanvas(new Point(xTic,0)).x, paddingTop)
        ..lineTo(toCanvas(new Point(xTic,0)).x, canvas.height-paddingBottom)
        ..stroke();
    }
    context.restore();
  }

  void drawTics() {
    var yTics = this.axes[AxisType.Y].tics;
    var xTics = this.axes[AxisType.X].tics;

    context.save();
    context.font = getFontString();
    context.fillStyle = "${Color.BLACK}";
    context.textAlign = "end";
    context.textBaseline = "middle";
    yTics.forEach((num yTic) {
      Point center = new Point(paddingLeft-fontSize.value/2, toCanvas(new Point(0, yTic)).y);
      context.fillText(yTic.toStringAsFixed(axes[AxisType.Y].fixedPrecision), center.x, center.y);
    });
    context.textAlign = "center";
    xTics.forEach((num xTic) {
      Point center = new Point(toCanvas(new Point(xTic, 0)).x, canvas.height - paddingBottom + fontSize.value);
      context.fillText(xTic.toStringAsFixed(axes[AxisType.X].fixedPrecision), center.x, center.y);
    });
    context.restore();
  }

  void drawAxisLabels() {
    context.save();
    context.translate(fontSize.value, canvas.height/2);
    context.strokeStyle = "${Color.BLACK}";
    context.fillStyle = "${Color.BLACK}";
    context.rotate(-math.PI/2);
    context.textAlign = "center";
    context.textBaseline="top";
    context.font = yAxisLabel.fontString;
    context.fillText(yAxisLabel.value, 0, 0);
    context.restore();
    
    context.save();
    context.font = xAxisLabel.fontString;
    context.strokeStyle = "${Color.BLACK}";
    context.fillStyle = "${Color.BLACK}";
    context.translate(graphWidth/2 + paddingLeft, canvas.height - fontSize.value);
    context.textAlign = "center";
    context.textBaseline = "bottom";
    context.fillText(xAxisLabel.value, 0, 0);
    context.restore();
  }

  bool validPointForText(Point p) {
    Point topLeft = toCanvas(rectangle.topLeft);
    Point bottomRight = toCanvas(rectangle.bottomRight);
    return p.x>topLeft.x+fontSize.value && p.x<bottomRight.x-fontSize.value  &&
            p.y<topLeft.y-fontSize.value && p.y>bottomRight.y+fontSize.value;
  }

  static bool ValidPointForCanvas(Point p)=>!(p.x.isNaN || p.x.isInfinite || p.y.isNaN || p.y.isInfinite);
  
}